from rest_framework import generics, status
from rest_framework.permissions import IsAuthenticated
from django.http import JsonResponse
from rest_framework.decorators import api_view, permission_classes
from clusterCore.models import c123_node, c123_host, c123_database, Task, c123_inventory, ClusterSizing
from clusterCore.enums import DBType, ClusterRoles, Inventory, Severity, Status, Schedule, RestoreOptions, CopyOptions, TaskState, Node_Size_Allowed
from clusterCore.api.serializer import c123_inventory_serializer, c123_DatabaseSerializer, c123_HostSerializer, c123_NodeSerializer, object_to_json, format_size, get_create_node_response
from django.utils import timezone
from clusterCore.exceptions import *
from .serializer import c123_CreateMySQLClusterSerializer, c123_AddMySQLStandbySerializer
from types import SimpleNamespace
from clusterCore.scheduler import schedule_backup, MISFIRE_GRACE_TIME_BACKUP
import json, re, pytz, traceback
from datetime import datetime
from dateutil.parser import parse
from clusterCore.tasks import allow_or_403, checkPathExists, check_space, get_cron_schedule, cleanup_logs, create_ssh_connection, \
remote_command, drop_mysql_node, create_mysql_connection, append_status_cmd, get_backup_description, export_mysql_node, export_mysql_database, \
get_avail_change_date
class MySQLStandbyAddView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_AddMySQLStandbySerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)
class MySQLClusterCreateView(generics.CreateAPIView):
    permission_classes = [IsAuthenticated]
    queryset = c123_node.objects.all()
    serializer_class = c123_CreateMySQLClusterSerializer
    def create(self, request, *args, **kwargs):
        return get_create_node_response(self, request, *args, **kwargs)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RestoreMySQLNode(request):
    """restore a mysql node from backup
    the call is made from client AJAX POST that contains the backup id"""
    
    bu_id = request.data.get('bu_id')
    # restore_option = 'this'/'latest'/'custom'
    restore_option = request.data.get('restore_option')
    restore_time = request.data.get('restore_time')
    autostart = request.data.get('autostart')
    api_call = request.data.get('rest_api')
    if api_call:
        backup = c123_inventory.objects.filter(name=bu_id, type = Inventory.BACKUP.value).first()
    else:
        backup = c123_inventory.objects.filter(pk=bu_id, type = Inventory.BACKUP.value).first()
    if not backup:
        raise APIException404("backup id {} not found".format(bu_id))
    node = backup.node
    allow_or_403(node)
    primary = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role=ClusterRoles.PRIMARY.value, ind_deleted=False).first()
    if not primary:
        #because it can not exist, or exist as deleted
        raise APIException410(f"primary of standby {standby.name} is not available")
    
    forbid_restore = Task.objects.filter(target_id = primary.id, task__in = ['restore_mysql_node_task', 'backup_mysql_node_task'], status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if forbid_restore:
        raise APIException409(f"Already runs backup/restore for primary {primary.name}")
    
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    
    path = backup.ex_params.get('backup_path')
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")
    
    primary.status = Status.RESTORING.value
    primary.save()
    if RestoreOptions.THIS.value in restore_option:
        date = backup.created_date.strftime("%Y-%m-%d %H:%M:%S")
    elif RestoreOptions.CUSTOM.value in restore_option and restore_time:
        # example date = '2021-12-30 09:08:33'
        # date = datetime.strptime(restore_time, "%Y-%m-%dT%H:%M:%S.%f%z").astimezone()
        # date = date.strftime("%Y-%m-%d %H:%M:%S")
        date = restore_time
    else:
        date = ''
    standbys_id = []
    standby_in_cluster = c123_node.objects.filter(type = DBType.MYSQL.value, cluster=primary.cluster, cluster_role=ClusterRoles.STANDBY.value, ind_deleted = False).exists()
    standbys = []
    if standby_in_cluster:
        cluster_standbys = c123_node.objects.filter(type = DBType.MYSQL.value, cluster = primary.cluster, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False)
        for standby in cluster_standbys:
            standbys_id.append(standby.id)
            standbys.append({"host": object_to_json(standby.host), "node": object_to_json(standby)})
    
    # serialize objects for the tasks
    backup = object_to_json(backup)
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data
    backup_host = c123_HostSerializer(app_host).data
    created_by = request.user
    if node.cluster_role == ClusterRoles.PRIMARY.value and c123_node.objects.filter(type = DBType.MYSQL.value, cluster = node.cluster, ind_deleted = False, severity = Severity.OK.value, cluster_role = ClusterRoles.STANDBY.value).exists():
         autostart = True
            
    tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'backup_data': backup, 'autostart': autostart, 'restore_option': restore_option, 'date': date, 'standbys_id': standbys_id, 'has_standby': standby_in_cluster, 'standbys': standbys, 'created_by': created_by.id})
    task = Task.objects.create(task='restore_mysql_node_task',
                            displayment='restore mysql node {}'.format(primary.name),
                            target_type=DBType.MYSQL.value,
                            target_id=primary.id,
                            org_id = primary.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMySQLUserExists(request):
    """check if a mysql user already exists"""
    
    primary_id = request.data.get("node_id")    
    username = request.data.get("username")
    primary = c123_node.objects.filter(id = primary_id, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException410(f"primary_id {primary_id} not available")
    allow_or_403(primary)
    conn = create_mysql_connection(primary)
    if isinstance(conn, tuple):
        if hasattr(conn[1], 'strerror'):
            raise APIException401(conn[1].strerror)
        raise APIException401(str(conn[1]))
    
    try:
        with conn:
            with conn.cursor() as cur:
                cur.execute("SELECT user FROM mysql.user WHERE user = %s AND host = '%.%.%.%'", (username,))
                result = cur.fetchone()
    except Exception as e:
        if hasattr(e, 'strerror'):
            raise APIException401(e.strerror)
        raise APIException401(str(e))
        
    if result:
        return JsonResponse({"exists": True}, status = 200)
    return JsonResponse({"exists": False}, status =  200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def UpgradeMySQLNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    target_version = request.data.get('target_version')
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    
    standbys_set = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = node.cluster)
    standbys_list = [{'node': object_to_json(s), 'host': object_to_json(s.host)} for s in standbys_set]
    
    c123_node.objects.filter(cluster = node.cluster, type = node.type, ind_deleted = False).update(status = Status.UPGRADING.value)
    
    tmp_data = json.dumps({'node_data': object_to_json(node), 'host_data': object_to_json(node.host), 'app_host': object_to_json(app_host), 'target_version': target_version, 'standbys': standbys_list, 'created_by': current_user.id})
    task = Task.objects.create(task='upgrade_mysql_node_task',
                                displayment='upgrading {} to mysql {}'.format(node.name, target_version),
                                target_type=DBType.MYSQL.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    return append_status_cmd(request, task)


@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def BackupMySQLNode(request):
    """backup for mysql node"""
    schedule = request.data.get('schedule')
    node_id = request.data.get('node_id')
    compress = request.data.get("compress")
    backup_path = request.data.get('backup_path')
    update_schedule = request.data.get("update_schedule")
    incremental = request.data.get('incremental')
    incremental_interval = request.data.get('incremental_interval')
    api_call = request.data.get('rest_api')
    from django.conf import settings
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    if api_call:
        mysql_node = c123_node.objects.filter(name=node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        mysql_node = c123_node.objects.filter(pk=node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not mysql_node:
        raise APIException410("node id {} not available".format(node_id))
    allow_or_403(mysql_node)
    primary = c123_node.objects.filter(type = DBType.MYSQL.value, cluster = mysql_node.cluster, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
         raise APIException404(f"primary of standby {mysql_node.name} not found")
      
    if schedule:
        client_timezone = pytz.timezone(request.data.get("client_timezone"))
        schedule_run_time_str = request.data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = request.data.get("days_of_week")
        days_of_month = request.data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
    if update_schedule and schedule: 
        backup = c123_inventory.objects.filter(node = primary, type = Inventory.BACKUP.value, is_sched = True).last()
        if backup:
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.ex_params['description'] = description
            backup.save()
            job_id = backup.ex_params.get("job_id")
            notification = f"{backup.node} backup updated to '{description}'"
            job = settings.scheduler.get_job(job_id=job_id)
            if job:
               job.reschedule(trigger=trigger)
            else:
                settings.scheduler.add_job(schedule_backup, trigger=trigger, replace_existing=True, kwargs={"task_id":  backup.ex_params["task_id"]}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
            Task.objects.filter(pk = backup.ex_params.get("binlog_task_id")).update(status = TaskState.PENDING.value)
            return JsonResponse({"notification": notification}, status=200)
        
        else:
            standbys = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, cluster = primary.cluster)
            if standbys.exists():
                for standby in standbys:
                    backup = c123_inventory.objects.filter(node = standby.pk, type = Inventory.BACKUP.value, is_sched = True).last()
                    if backup:
                        backup.ex_params['description'] = description
                        backup.ex_params['schedule_run_time'] = schedule_run_time_str
                        backup.ex_params['days_of_month'] = days_of_month
                        backup.ex_params['days_of_week'] = days_of_week
                        backup.save()
                        notification = f"{backup.node} backup updated to '{description}'"
                        job_id = backup.ex_params.get("job_id")
                        job = settings.scheduler.get_job(job_id=job_id)
                        if job:
                           job.reschedule(trigger=trigger)
                        else:
                            settings.scheduler.add_job(schedule_backup, trigger=trigger, kwargs={"task_id": backup.ex_params["task_id"]}, replace_existing=True, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP, id=job_id)
                        Task.objects.filter(pk = backup.ex_params.get("binlog_task_id")).update(status = TaskState.PENDING.value)
                        return JsonResponse({"notification": notification}, status=200)
                           
                
                return JsonResponse({"notification": "no schedule backup found"}, status = 200)            
            else:
               return JsonResponse({"notification": "no schedule backup found"}, status = 200)
    serializer = c123_inventory_serializer(data = request.data, partial = True)
    if serializer.is_valid():
    # create the task signature
        check_space(node = mysql_node, path=backup_path, source_path=mysql_node.dir_123Cluster)
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MYSQL.value)
        keep_days = getattr(size_scheme, f'{mysql_node.sizing.lower()}_backup_keep_days', None)
        if schedule:
            existing_backup = c123_inventory.objects.filter(node = mysql_node, type = Inventory.BACKUP.value, is_sched = True)
            if existing_backup.exists():
                raise APIException409("Schedule backup already exists.\nCan be only one schedule backup for cluster.")
            
            if mysql_node.cluster_role == ClusterRoles.PRIMARY.value:
                standbys = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, cluster = mysql_node.cluster)
            else:
                standbys = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, cluster = mysql_node.cluster).exclude(pk = mysql_node.pk)
            
            if standbys.exists():
                for standby in standbys:
                    existing_backup = c123_inventory.objects.filter(node = standby, type = Inventory.BACKUP.value, is_sched = True)
                    if existing_backup.exists():
                        raise APIException409("Schedule backup already exists.\nCan be only one schedule backup for cluster.")
            backup = serializer.create(serializer.validated_data)
            backup.node = mysql_node
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.MYSQL.value
            backup.ex_params["node_version"] = mysql_node.version
            backup.ex_params["incremental"] = incremental
            backup.ex_params['description'] = description
            backup.ex_params['schedule_run_time'] = schedule_run_time_str
            backup.ex_params['days_of_month'] = days_of_month
            backup.ex_params['days_of_week'] = days_of_week
            backup.is_sched = schedule
            if keep_days:
                backup.keep_days = keep_days
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["modified_date"] = ''
            notification =  "{} backup {}".format(backup.node, description)
            job_id = f"{notification} ({client_timezone})"
            backup.ex_params["job_id"] = job_id
            backup.ex_params["root_pwd"] = mysql_node.ex_params.get('mysql_root_password')
            backup.ex_params["backup_path"] = f'{backup_path}/{mysql_node.host.name}_{mysql_node.port}/{"incremental" if incremental else "schedule_full"}'
            backup.ex_params["start_date"] = ''
            backup.ex_params['incremental_interval'] = incremental_interval
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(mysql_node.name, now)
            backup.save()
            # serialize objects for the tasks
            node_data = object_to_json(mysql_node)
            host_data = c123_HostSerializer(mysql_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user
            
            tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'backup_data': object_to_json(backup), 'created_by': created_by.id, 'compress': compress})
            task = Task.objects.create(task='backup_mysql_node_task',
                                            displayment='backup {}'.format(mysql_node.name), 
                                            target_type=DBType.MYSQL.value, 
                                            target_id=mysql_node.id, 
                                            org_id = mysql_node.org_id,
                                            schedule=Schedule.SCHEDULED.value, 
                                            status=TaskState.ON_DEMAND.value,
                                            parameters=tmp_data,
                                            created_by=created_by)
               
            backup.ex_params['task_id'] = task.pk
            backup.save()
            settings.scheduler.add_job(schedule_backup, trigger=trigger, kwargs={"task_id": task.pk}, replace_existing=True, id=job_id , misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        else:
            existing_backup = Task.objects.filter(target_id = mysql_node.id, task = 'backup_mysql_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
            if existing_backup:
                raise APIException409(f"Already runs backup for {mysql_node.name}")
            backup = serializer.create(serializer.validated_data)
            backup.node = mysql_node
            backup.type = Inventory.BACKUP.value
            backup.created_by = request.user
            backup.ex_params = {"size": 0}
            backup.ex_params["belong_to"] = DBType.MYSQL.value
            backup.ex_params["node_version"] = mysql_node.version
            backup.is_sched = schedule
            if keep_days:
                backup.keep_days = keep_days
            backup.ex_params["duration"] = 0
            backup.ex_params["compress"] = compress
            backup.ex_params["root_pwd"] = mysql_node.ex_params.get('mysql_root_password')
            backup.ex_params["modified_date"] = ''
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            backup.name = "{}_{}".format(mysql_node.name, now)
            backup.ex_params["backup_path"] = f'{backup_path}/{mysql_node.host.name}_{mysql_node.port}/{now}'
            backup.save()   
            notification = f"backup of {mysql_node.name}"
            # serialize objects for the tasks
            node_data = object_to_json(mysql_node)
            host_data = c123_HostSerializer(mysql_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            created_by = request.user
            tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'backup_data': object_to_json(backup), 'created_by': created_by.id, 'compress': compress})
            task = Task.objects.create(task='backup_mysql_node_task',
                                            displayment='backup {}'.format(mysql_node.name), 
                                            target_type=DBType.MYSQL.value, 
                                            target_id=mysql_node.id,
                                            org_id = mysql_node.org_id, 
                                            schedule=Schedule.NOW.value, 
                                            status=TaskState.PENDING.value,
                                            parameters=tmp_data,
                                            created_by=created_by)
            
            backup.ex_params['task_id'] = task.pk
            backup.save()
    else:
        print("invalid form")
        print(serializer.errors)        
        return JsonResponse({'error': serializer.errors}, status=404)
    return append_status_cmd(request, task, notification)


api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def GetMySQLSettings(request, id):
    node = c123_node.objects.filter(type = DBType.MYSQL.value, pk = id,  ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(id))
    
    conn = create_mysql_connection(node)
    if isinstance(conn, tuple):
        if hasattr(conn[1], 'strerror'):
            raise APIException401(conn[1].strerror)
        raise APIException401(conn[1])
    
    try:
        with conn:
            with conn.cursor(dictionary = True) as cur:
                cur.execute("SELECT v.VARIABLE_NAME AS name, v.VARIABLE_SOURCE AS context, v.VARIABLE_PATH AS source, gv.VARIABLE_VALUE AS setting, CASE WHEN v.VARIABLE_PATH = '' OR v.VARIABLE_SOURCE = 'PERSISTED' THEN 'NO' ELSE 'YES' END AS read_only, 'default' as color FROM performance_schema.variables_info v LEFT JOIN performance_schema.global_variables gv  ON v.VARIABLE_NAME = gv.VARIABLE_NAME WHERE v.VARIABLE_SOURCE NOT IN ('COMPILED', 'COMMAND_LINE') AND v.VARIABLE_NAME NOT IN ('datadir', 'performance_schema');")
                settings = cur.fetchall()
                return JsonResponse(settings, status = 200, safe = False)

    except Exception as e:
        raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))
     

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def EditMySQLSettings(request):
   
    try:
        node_id = request.data.get('node_id')
        parameters = request.data.get('parameters')
        created_by = request.user
        node = c123_node.objects.filter(type = DBType.MYSQL.value, pk = node_id,  ind_deleted = False).first()
        if not node:
            raise APIException410("node id {} not available".format(node_id))

        tmp_data = json.dumps({'host_data': object_to_json(node.host), 'node_data': object_to_json(node), 'parameters': parameters, 'created_by': created_by.id})
        task = Task.objects.create(task='mysql_parameters_apply_task',
                                displayment=f'applying parameters to node {node.name}',
                                target_type=DBType.MYSQL.value,
                                target_id=node.pk,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
        return append_status_cmd(request, task)

    except Exception as e:
        print(traceback.format_exc())
        raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RemoveMySQLSchedule(request):
    node_id = request.data.get("node_id")
    primary = c123_node.objects.filter(pk=node_id, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value).first()
    if not primary:
        raise APIException404("primary id {} not found".format(node_id))
    allow_or_403(primary)
    from django.conf import settings
    backup = c123_inventory.objects.filter(node = primary, type = Inventory.BACKUP.value, is_sched = True).last()
    if backup:
        Task.objects.filter(pk = backup.ex_params.get("binlog_task_id")).update(status = TaskState.ON_DEMAND.value)
        job_id = backup.ex_params.get("job_id")
        job = settings.scheduler.get_job(job_id=job_id)
        if job:
            settings.scheduler.remove_job(job_id=job_id)
            notification = f"{job_id} removed"
        # job = settings.scheduler.get_job(f'mysqlbinlog_dump_task_{primary.cluster.pk}')
        # if job:
        #     settings.scheduler.remove_job(f'mysqlbinlog_dump_task_{primary.cluster.pk}')
            return JsonResponse({"notification": notification}, status=200)
        
        else:
            standbys = c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, cluster = primary.pk)
            if standbys.exists():
                for standby in standbys:
                    backup = c123_inventory.objects.filter(node = standby, type = Inventory.BACKUP.value, is_sched = True).last()
                    if backup:
                        Task.objects.filter(pk = backup.ex_params.get("binlog_task_id")).update(status = TaskState.ON_DEMAND.value)
                        job_id = backup.ex_params.get("job_id")
                        job = settings.scheduler.get_job(job_id=job_id)
                        if job:
                            settings.scheduler.remove_job(job_id=job_id)
                            notification = f"{job_id} removed"
                            # job = settings.scheduler.get_job(f'mysqlbinlog_dump_task_{primary.cluster.pk}')
                            # if job:
                            #     settings.scheduler.remove_job(f'mysqlbinlog_dump_task_{primary.cluster.pk}')
                            return JsonResponse({"notification": notification}, status=200)
                
                return JsonResponse({"notification": "no schedule backup found"}, status = 200)            
            else:
               return JsonResponse({"notification": "no schedule backup found"}, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def StopMySQLNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node:
        raise APIException410("node id {} not available".format(node_id))
    node.ind_started = False
    node.status = Status.STOPPING.value
    node.save()
    tmp_data = json.dumps({'node_id': node.pk, 'host_data': object_to_json(node.host), 'created_by': current_user.id})
    task = Task.objects.create(task='stop_mysql_node_task',
                                    displayment='stopping node {}'.format(node.name),
                                    target_type=DBType.MYSQL.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def StartMySQLNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    current_user = request.user
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node:
        raise APIException404("node id {} not found".format(node_id))
    if node.ind_started:
         node.status = Status.RESTARTING.value
         node.save()
         tmp_data = json.dumps({'node_id': node.id, 'host_data': object_to_json(node.host), 'created_by': current_user.id})
         task = Task.objects.create(task='restart_mysql_node_task',
                                    displayment='restarting node {}'.format(node.name),
                                    target_type=DBType.MYSQL.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    else:
        node.ind_started = True
        node.status = Status.STARTING.value
        node.save()
        tmp_data = json.dumps({'node_id': node.pk, 'host_data': object_to_json(node.host), 'created_by': current_user.id})
        task = Task.objects.create(task='start_mysql_node_task',
                                    displayment='starting node {}'.format(node.name),
                                    target_type=DBType.MYSQL.value,
                                    target_id=node.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['GET'])
@permission_classes((IsAuthenticated, ))
def MySQLDetails(request, id):
    node = c123_node.objects.filter(pk = id, type = DBType.MYSQL.value).first()
    if not node:
        raise APIException404("node id {} not found".format(id))
    response = {'node': c123_NodeSerializer(node).data}
    auto_backups = c123_inventory.objects.filter(node=node, type = Inventory.BACKUP.value, is_sched=True)
    if auto_backups.count() > 0:
        if auto_backups.last().status:
            response["auto_backup"] = {
                    'status': "Automatic backup", 'class': "green-star", "detail": "Automatic backup configured"}
        else:
            response["auto_backup"] = {
                    'status': "Automatic backup", 'class': "red-green-star", "detail": "Automatic backup is configured, but the last run reported errors"}
                
    else:
        response["auto_backup"]  = {
                'status': "No automatic backup", 'class': "red-star", "detail" : "Automatic backup is not configured"}
    if node.cluster_role == ClusterRoles.PRIMARY.value:
        primary_backups = c123_inventory_serializer(c123_inventory.objects.filter(node_id = node.id, type = Inventory.BACKUP.value), many = True)
        response["backups"] = primary_backups.data
        primary_exports = c123_inventory_serializer(c123_inventory.objects.filter(node_id = node.id, type = Inventory.EXPORT.value), many = True)
        response["exports"] = primary_exports.data
        standbys = c123_NodeSerializer(c123_node.objects.filter(type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = node.cluster), many = True)
        response["nodes"] = standbys.data
        for i, standby in enumerate(response["nodes"]):
            backups = c123_inventory_serializer(c123_inventory.objects.filter(node_id = standby["id"], type = Inventory.BACKUP.value), many = True)
            response["nodes"][i]["backups"] = backups.data
            exports = c123_inventory_serializer(c123_inventory.objects.filter(node_id = standby["id"], type = Inventory.EXPORT.value), many = True)
            response["nodes"][i]["exports"] = exports.data
    else:
        standbys = c123_NodeSerializer(c123_node.objects.filter(type = DBType.MYSQL.value, cluster = node.cluster, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False).exclude(pk = node.pk), many = True)
        response["nodes"] = standbys.data
    
 
            
    cs = ClusterSizing.objects.filter(db_type = DBType.MYSQL.value).first()
    size_scheme = {
            Node_Size_Allowed.SMALL.name: {
                'cpu': cs.small_cpu,
                'disk': cs.small_disk,
                'memory': cs.small_memory,
                'backup': cs.small_backup_keep_days,
                'export': cs.small_export_keep_days,
            },
            Node_Size_Allowed.MEDIUM.name: {
                'cpu': cs.medium_cpu,
                'disk': cs.medium_disk,
                'memory': cs.medium_memory,
                'backup': cs.medium_backup_keep_days,
                'export': cs.medium_export_keep_days,
            },
            Node_Size_Allowed.LARGE.name: {
                'cpu': cs.large_cpu,
                'disk': cs.large_disk,
                'memory': cs.large_memory,
                'backup': cs.large_backup_keep_days,
                'export': cs.large_export_keep_days,
            },
            Node_Size_Allowed.VERY_LARGE.name: {
                'cpu': cs.very_large_cpu,
                'disk': cs.very_large_disk,
                'memory': cs.very_large_memory,
                'backup': cs.very_large_backup_keep_days,
                'export': cs.very_large_export_keep_days,
            }
        }
    response["avail_schemes"] = size_scheme
    response["databases"] = c123_DatabaseSerializer(c123_database.objects.filter(node = node, ind_deleted = False), many = True).data


    conn = create_mysql_connection(node)
    if isinstance(conn, tuple):
        if hasattr(conn[1], 'strerror'):
            error = conn[1].strerror
        else:
            error = str(conn[1])
        
        response["detail"] = error
        response["ind_error"] = c123_node.objects.filter(type = DBType.MYSQL.value, cluster = node.cluster, ind_deleted = False, severity__in=[Severity.WARNING.value, Severity.NOT_OK.value]).exists()  or c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).count() < 2
        response['node'] = c123_NodeSerializer(c123_node.objects.filter(pk = id,  type = DBType.MYSQL.value).first()).data
        response['node']['ex_params']['avail_change'] =  get_avail_change_date(node)
        return JsonResponse(response, status = 401)
    else:
        response["ind_error"] = c123_node.objects.filter(type = DBType.MYSQL.value, cluster = node.cluster, ind_deleted = False, severity__in=[Severity.WARNING.value, Severity.NOT_OK.value]).exists() or c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).count() < 2
        response['node'] = c123_NodeSerializer(c123_node.objects.filter(pk = id,  type = DBType.MYSQL.value).first()).data
        response['node']['ex_params']['avail_change'] =  get_avail_change_date(node)
    try:
        with conn:
            with conn.cursor() as cur:
                cur.execute("SELECT VERSION()")
                version = cur.fetchone()[0]
                version = version.split('-')[0].split(".")
                response["version"] = ".".join(version)
                node.version = f'{version[0]}.{version[1]}'
                node.save()
                cur.execute("SELECT DATABASE()")
                response['db_id'] = cur.fetchone()[0]
                cur.execute("SHOW VARIABLES LIKE 'sql_log_bin'")
                response['logging'] = cur.fetchone()[0]
                cur.execute("SHOW VARIABLES LIKE 'read_only'")
                response['read_only'] = cur.fetchone()[0]
    except Exception as e:
        if hasattr(e, 'strerror'):
            error = e.strerror
        else:
            error = str(e)
        response["detail"] = error
        return JsonResponse(response, status = 403)
    
    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            error = client[1].strerror
        else:
            error = str(client[1])
        
        response["detail"] = error
        return JsonResponse(response, status = 401)

    command = f'sudo journalctl -u mysql.service -n 51 -o json | tac'
    jsn = json.loads(remote_command(client, command, True), strict=False)
    if(jsn['stderr'] != '' and "error" in jsn["stderr"].lower()):
        error = jsn["stderr"]
        print(error)
        response["error"] = error
        return JsonResponse(response, status=404)
    
    if jsn['stdout'] == '':
         command = 'sudo journalctl -u mysqld.service -r -n 51 -o json | tac'
         jsn = json.loads(remote_command(client, command, True), strict=False)
   
    last_20_logs = [json.loads(line, strict = False) for line in jsn['stdout'].split('\n') if line != '']
    response['history'] = last_20_logs

    command = f"sudo du -bs  '{node.dir_123Cluster}' | awk '{{print $1}}'"
    jsn = json.loads(remote_command(client, command, True), strict=False)
    if(jsn['stderr'] != ''):
        print(jsn['stderr'])
        response["detail"] = jsn['stderr']
        return JsonResponse(response, status=404)
    node.size = round(float(jsn['stdout']) / 1024, 2)
    node.save()
    response["display_size"] = format_size(node.size)

    jsn = json.loads(remote_command(client, f"""sudo netstat -tunlp | grep '{node.port}' | awk '{{print $7}}' | awk -F'/' '{{print $1}}'  | sort -u | xargs -I{{}} bash -c 'awk "/^VmRSS:/ {{print \$2}}" /proc/{{}}/status && awk "{{print (\$14+\$15)/100}}" /proc/{{}}/stat' | awk '{{total_mem+=$1; total_cpu+=$2}} END {{print total_cpu, total_mem}}'""", True), strict=False)
    if jsn['stdout'] != '':
        print(jsn['stdout'])
    
    if jsn['stderr'] != '':
        response["detail"] = jsn['stderr']
        return JsonResponse(response, status = 404)
    
    measurements = jsn['stdout'].split(' ')
    response["cpu_percent"] = measurements[0]
    response["display_ram"] = format_size(measurements[1])
    
    return JsonResponse(response, status = 200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def MySQLClusterStatus(request, id):
    primary = c123_node.objects.filter(pk = id, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException410(f"primary id {id} not available")
    standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MYSQL.value, ind_deleted = False, cluster = primary.cluster)
    if standbys.count() == 0:
        raise APIException403(f"Primary {primary.name} has no available standbys.\nCannot continue.")
    status = []
    for standby in standbys:
            conn = create_mysql_connection(standby)
            if isinstance(conn, tuple):
                if hasattr(conn[1], 'strerror'):
                    print(conn[1].strerror)
                else:
                    print(str(conn[1]))
                continue
        
            try:
                with conn:
                    with conn.cursor(dictionary=True) as cur:
                        cur.execute(f"SHOW {'REPLICA' if float(standby.version) > 5.7 else 'SLAVE'} STATUS")
                        entry = cur.fetchone()
                        entry[f"{'Replica' if float(standby.version) > 5.7 else 'Slave'}_Host"] = standby.name
                        status.append(entry)
            except Exception as e:
                if hasattr(e, 'strerror'):
                    print(e.strerror)
                else:
                    print(str(e))
                continue

    return JsonResponse(status, status = 200, safe = False)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ExportMySQLNode(request):
    """export for mysql node"""

    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')

    if not schedule:
        task = export_mysql_node(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            node = c123_node.objects.filter(type = DBType.MYSQL.value, name=node_id, ind_deleted = False).first()
        else:
            node = c123_node.objects.filter(type = DBType.MYSQL.value, pk=node_id, ind_deleted = False).first()
        if not node:
            raise APIException410("node id {} not available".format(node_id))
        job_id =  f"export_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_mysql_node, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export {node.name} {description} started"}, status = 200)
    
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ExportMySQLDatabase(request):
    """export for database"""
    data = request.data
    data['created_by'] = request.user.id
    schedule = data.get('schedule')
    api_call = data.get('api_call')
    node_id = data.get('node_id')
    db_id = data.get('db_id')

    if not schedule:
        task = export_mysql_database(data)
        return append_status_cmd(request, task)
    else:
        from django.conf import settings
        client_timezone = pytz.timezone(data.get("client_timezone"))
        schedule_run_time_str = data.get('schedule_run_time')
        schedule_run_time = parse(schedule_run_time_str).astimezone(client_timezone)
        client_hours = schedule_run_time.hour
        client_minutes = schedule_run_time.minute
        client_seconds = schedule_run_time.second
        days_of_week = data.get("days_of_week")
        days_of_month = data.get("days_of_month")
        if days_of_month:
            int_days_of_month = [parse(d).astimezone(client_timezone).day for d in days_of_month]
        else:
            int_days_of_month = None
        description = get_backup_description(client_hours, client_minutes, client_seconds, days_of_week, int_days_of_month)
        localized_datetime = datetime.now(client_timezone).replace(hour=client_hours, minute=client_minutes, second=client_seconds).astimezone(pytz.timezone(settings.TIME_ZONE))
        hours = localized_datetime.hour
        minutes = localized_datetime.minute
        seconds = localized_datetime.second
        trigger = get_cron_schedule(hours, minutes, seconds, days_of_week, int_days_of_month)
        if api_call:
            db = c123_database.objects.filter(type = DBType.MYSQL.value, name=db_id, node__name = node_id, ind_deleted = False).first()
        else:
            db = c123_database.objects.filter(type = DBType.MYSQL.value, pk=db_id, ind_deleted = False).first()
        if not db:
            raise APIException410("database id {} not available".format(db_id))
        node = db.node
        job_id =  f"export_{db.name}_{node.name}"
        scheduler = settings.scheduler
        export_job = next((i for i in scheduler.get_jobs() if job_id in i.id), None)
        if export_job:
            export_job.reschedule(trigger=trigger)
        else:
             settings.scheduler.add_job(export_mysql_database, replace_existing=True, trigger=trigger,  kwargs={"data": data}, id=job_id, misfire_grace_time=MISFIRE_GRACE_TIME_BACKUP)
        return JsonResponse({"notification": f"export db {db.name} of {node.name} {description} started"}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def RestoreMySQLDatabase(request):
    """restore a mysql db from export
    the call is made from client AJAX POST that contains the backup id"""
    export_id = request.data.get('export_id')
    api_call = request.data.get('rest_api')
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    autostart = True if request.data.get('autostart') == 'true' else False
    if api_call:
        export = c123_inventory.objects.filter(name=export_id, type = Inventory.EXPORT.value).first()
    else:
        export = c123_inventory.objects.filter(pk=export_id, type = Inventory.EXPORT.value).first()
    if not export:
        raise APIException404("export id {} not found".format(export_id))
    
    path = export.ex_params.get('backup_path')
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")
    
    db = export.database
    node = db.node
    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
          raise APIException404("primary not found")
    allow_or_403(primary)
    db.status = Status.RESTORING.value
    db.save()
    backup_host = c123_HostSerializer(app_host).data
    host_data = c123_HostSerializer(primary.host).data
    db_data = object_to_json(db)
    export = object_to_json(export)
    node_data = object_to_json(primary)
    created_by = request.user
    tmp_data = json.dumps({'db_data': db_data, 'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'backup_data': export, 'backup_type': 'dump', 'autostart': autostart, 'created_by': created_by.id})
    task = Task.objects.create(task='mysqldump_restore_task',
                            displayment='restore mysql database {}'.format(db.name),
                            target_type=DBType.MYSQL.value,
                            target_id=db.id,
                            org_id = node.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def ImportMySQLNode(request):
    """restore a mysql node from export
    the call is made from client AJAX POST that contains the backup id"""
    export_id = request.data.get('export_id')
    api_call = request.data.get('rest_api')
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")  
    if api_call:
        export = c123_inventory.objects.filter(name=export_id, type = Inventory.EXPORT.value).first()
    else:
        export = c123_inventory.objects.filter(pk=export_id, type = Inventory.EXPORT.value).first()
    if not export:
        raise APIException404(f"export id {export_id} not found")
    path = export.ex_params.get('backup_path')
    path_exists = checkPathExists(app_host, path)
    if not path_exists:
        raise APIException404(f"{path}: no such file or directory")

    backup_host = c123_HostSerializer(app_host).data
    node = export.node

    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
          raise APIException404("primary not found")
    allow_or_403(primary)
    node.status = Status.RESTORING.value
    node.save()
    standby_queryset = c123_node.objects.filter(cluster = primary.cluster, type = DBType.MYSQL.value,  cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False)
    standbys = []
    for standby in standby_queryset:
        standbys.append({"node": object_to_json(standby), "host": object_to_json(standby.host)})
    # serialize objects for the tasks
    export = object_to_json(export)
    node_data = object_to_json(primary)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user
    tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host, 'standbys': standbys, 'backup_data': export, 'backup_type': 'dumpall', 'created_by': created_by.id})
    task = Task.objects.create(task='mysqldump_restore_task',
                            displayment='restore mysql node {}'.format(primary.name),
                            target_type=DBType.MYSQL.value,
                            target_id=primary.id,
                            org_id = primary.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMySQLVersionExists(request):
    """check if a postgres version package already exists"""
    host_name = request.data.get("host")
    username = request.data.get("os_username")
    password = request.data.get("os_password")
    private_key = request.data.get("private_key")
    ssh_port = request.data.get("ssh_port")
    version = request.data.get("software_version")
        
    host = SimpleNamespace( **{
        "name": host_name, 
        "username": username, 
        "password": password,
        "private_key": private_key,
        "ssh_port": ssh_port
    })
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    
    jsn = json.loads(remote_command(client, f'rpm -qa | grep -i "mysql-wsrep-{version}"', True))
    if(jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower()):
        print(jsn['stderr'])
        if 'not found' in jsn['stderr']:
            jsn = json.loads(remote_command(client, f'dpkg -l | grep -i "mysql-wsrep-{version}"', True))
            
            if(jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower()):                
                return JsonResponse({'error': jsn['stderr']}, status=404)
            elif(jsn['stdout'] != ''):
                return JsonResponse({'exists': True}, status=200)
    if(jsn['stdout'] != ''):
        return JsonResponse({'exists': True}, status=200)
        
    return JsonResponse({'exists': False}, status=200)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMySQLDatabaseName(request):
    
    node_id = request.data.get('node_id')
    db_name = request.data.get('name')
    node = c123_node.objects.filter(pk=node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node:
        raise APIException410(f"node id {node_id} not available")
    
    if(re.match(r'^[a-zA-Z0-9_]{1,64}$', db_name)):    
        conn = create_mysql_connection(node)
        if isinstance(conn, tuple):
            raise APIException401(conn[1].strerror if hasattr(conn[1], 'strerror') else str(conn[1]))
        try:
            with conn:
                with conn.cursor() as cur:
                    cur.execute("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = %s", (db_name.strip(),))
                    res = cur.fetchone()
                    if res:
                        return JsonResponse({'error': "already exists"}, status=409)
                    return JsonResponse({'response': "name available"}, status=200)
        except Exception as e:
            raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))
    else:
        return JsonResponse({'error': "invalid name"}, status=406)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def AddMySQLDatabase(request):
    current_user = request.user
    db_name = request.data.get('db_name')
    if not db_name or db_name.isspace():
          raise APIException406("database name cannot be blank")
    db_name = ''.join(db_name.split())
    node_id = request.data.get('node_id')
    api_call = request.data.get('rest_api')
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()

    if not node:
        raise APIException410("node id {} not available".format(node_id))
    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException404("no found primary for node {}".format(node.name))
    allow_or_403(primary)
    host = primary.host
    auto_delete = request.data.get('auto_delete')
    auto_delete_days = request.data.get('auto_delete_days')
    
    existing_dbs = c123_database.objects.filter(name=db_name, node=primary)
    if existing_dbs.exists():
        undeleted = existing_dbs.filter(ind_deleted = False)
        if undeleted.exists():
            db = undeleted.first()
        else:
            db = existing_dbs.first()
            db.ind_deleted = False
        
        db.created_by = current_user
        db.last_update_by = current_user
        db.status = Status.CREATING.value
        db.database_user = primary.database_user
        db.database_pwd = primary.database_pwd
        db.type = DBType.MYSQL.value
        db.save()
    else:    
        db = c123_database.objects.create(name=db_name,
            type=DBType.MYSQL.value,
            created_by=current_user,
            last_update_by=current_user,
            status='being created',
            severity=Severity.NOT_OK.value,
            database_user=primary.database_user,
            database_pwd=primary.database_pwd,
            node = primary)
    tmp_data = json.dumps({'node_data': object_to_json(primary), 'host_data': object_to_json(host), 'db': object_to_json(db), 'auto_delete': auto_delete, 'created_by': current_user.id})
    task = Task.objects.create(task='create_mysql_database_task',
                                    displayment='creating mysql db {}'.format(db_name),
                                    target_type=DBType.MYSQL.value,
                                    target_id=db.id,
                                    org_id = node.org_id,
                                    status=TaskState.PENDING.value,
                                    schedule=Schedule.NOW.value,
                                    parameters=tmp_data,
                                    created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropMySQLNode(request):
    node_id = request.data.get("node_id")
    api_call = request.data.get('rest_api')
    final_backup = request.data.get('final_backup')
    if api_call:
        node = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()

    if not node:
        raise APIException410(f"node id {node_id} not available")
    
    task = drop_mysql_node(node, final_backup)
         
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def DropMySQLDatabase(request):
           
    db_id = request.data.get("db_id")
    node_id = request.data.get("node_id")
    api_call = request.data.get("rest_api")
    if api_call:
        db = c123_database.objects.filter(name = db_id, node__name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        db = c123_database.objects.filter(pk = db_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not db:
        raise APIException410("database id {} not available".format(db_id))
    node = db.node
    
    primary = c123_node.objects.filter(cluster = node.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException404("no found primary for node {}".format(node.name))
    db.status = Status.DROPPING.value
    db.save()
    cleanup_logs(db.pk, db.type)
    node_data = object_to_json(primary)
    db_data = object_to_json(db)
    host_data = c123_HostSerializer(primary.host).data
    created_by = request.user
    
    tmp_data = json.dumps({'node_data': node_data, 'db_data': db_data, 'host_data': host_data, 'created_by': created_by.id})               
    task = Task.objects.create(task='drop_mysql_database_task',
                            displayment='drop database {}'.format(db.name),
                            target_type=DBType.MYSQL.value,
                            target_id=db.pk,
                            org_id = primary.org_id,
                            schedule=Schedule.NOW.value,
                            status=TaskState.PENDING.value,
                            parameters=tmp_data,
                            created_by=created_by)
         
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CopyMySQLDatabase(request):
    db_id = request.data.get("db_id")
    source_id = request.data.get("source_id")
    node_id = request.data.get("node_id")
    current_user = request.user
    db_option = request.data.get("option")
    #copy_option = 'one_time' or 'live'
    copy_option = request.data.get("copy_option")
    copy_name = request.data.get("copy_name")
    copy_users = request.data.get('copy_users')
    db_exists = request.data.get("db_exists")
    api_call = request.data.get('rest_api')
    if api_call:
        node_target = c123_node.objects.filter(name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        node_target = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node_target:
        raise APIException404("target node id {} not found".format(node_id))
    if api_call:
        node_source = c123_node.objects.filter(name = source_id, type = DBType.MYSQL.value, ind_deleted =False).first()
    else:
        node_source = c123_node.objects.filter(pk = source_id, type = DBType.MYSQL.value, ind_deleted =False).first()     
    if not node_source:
        raise APIException404("source node id {} not found".format(source_id))
    if api_call:
        db_target = c123_database.objects.filter(name = db_id, node = node_source, type = DBType.MYSQL.value, ind_deleted = False).first()
    else:
        db_target = c123_database.objects.filter(pk = db_id, type = DBType.MYSQL.value, ind_deleted = False).first()
        
    if not db_target:
        raise APIException410("target database id {} not available".format(db_id))
    
    primary_target = c123_node.objects.filter(cluster = node_target.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary_target:
        raise APIException404("no found primary for target node {}".format(node_target.name))
    allow_or_403(primary_target)
    primary_source = c123_node.objects.filter(cluster = node_source.cluster, type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary_source:
        raise APIException404("no found primary for source node {}".format(node_source.name))
    allow_or_403(primary_source)
    #source db
    db_source = c123_database.objects.get(pk=db_target.pk)
    if(db_option):
        # here we do for existing database
        match db_option:
            case CopyOptions.OVERRIDE.value:
                copy_name = db_target.name
                db_target.ind_started = False
                db_target.node = primary_target
                db_target.database_user = primary_target.database_user
                db_target.database_pwd = primary_target.database_pwd
                db_target.status = Status.REPLICATING.value
                db_target.save()
            case CopyOptions.COPY.value:
                existing_dbs = c123_database.objects.filter(node = primary_target, name = copy_name)
                if not existing_dbs.exists():
                        db_target = c123_database.objects.create(node = primary_target, name = copy_name,
                        type = db_source.type, size = db_source.size,
                        database_user = primary_target.database_user, database_pwd = primary_target.database_pwd,
                        created_by = current_user, last_update_by = current_user,
                        severity = Severity.NOT_OK.value, status = Status.REPLICATING.value
                        )
                else:
                    undeleted = existing_dbs.filter(ind_deleted = False)
                    if undeleted.exists():
                        db_target = undeleted.first()
                    else:
                        db_target = existing_dbs.first()
                        db_target.ind_deleted = False
                        
                    db_target.type = DBType.MYSQL.value
                    db_target.status = Status.REPLICATING.value
                    db_target.save()
                
                db_source.ind_started = False
                db_source.status = Status.COPYING.value
                db_source.save()
    else:
        #here we do for new database
        db_source.status =  Status.COPYING.value
        db_source.ind_started = False
        db_source.save()
        db_target = c123_database.objects.create(node = primary_target, name = db_source.name,
        type = db_source.type, size = db_source.size,
        database_user = primary_target.database_user, database_pwd = primary_target.database_pwd,
        created_by = current_user, last_update_by = current_user,
        severity = Severity.NOT_OK.value, status = Status.REPLICATING.value
        )
    tmp_data = json.dumps({'db_source': object_to_json(db_source), 'db_target' : object_to_json(db_target), 'node_source': object_to_json(primary_source), 'node_target': object_to_json(primary_target), 'host_source': object_to_json(primary_source.host), 'host_target': object_to_json(primary_target.host), 'db_option': db_option, 'db_exists': db_exists, 'copy_option': copy_option, 'copy_name': copy_name, 'copy_users': copy_users, 'created_by': current_user.id})
    task = Task.objects.create(task='copy_mysql_database_task',
                                displayment=f'copy database {db_source.name} into {node_target.name}',
                                target_type=DBType.MYSQL.value,
                                target_id=primary_target.pk,
                                org_id = primary_target.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    return append_status_cmd(request, task)
@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def CheckMySQLDatabaseSize(request):
    
    node_id = request.data.get('node_id')
    db_name = request.data.get("db_name")
    node = c123_node.objects.filter(pk = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
    if not node:
        raise APIException410(f"node id {node_id} not available")
    
    conn = create_mysql_connection(node)
    if isinstance(conn, tuple):
        if hasattr(conn[1], 'strerror'):
            raise APIException401(conn[1].strerror)
        raise APIException401(conn[1])
    
    try:
        with conn:
            with conn.cursor() as cur:
                if db_name:
                    cur.execute("SELECT sum(data_length + index_length)/1024 FROM information_schema.TABLES WHERE table_schema = %s GROUP BY table_schema", (db_name.strip(),))
                else:
                    cur.execute("SELECT sum(data_length + index_length)/1024 FROM information_schema.TABLES")
                res = cur.fetchone()
                db_size = res[0] if res and res[0] else 0.0
    except Exception as e:
        raise APIException401(e.strerror if hasattr(e, 'strerror') else str(e))    
    
    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])
    
    command = 'df -P . | tail -1 | awk \'{print $4}\''
    jsn = json.loads(remote_command(client, command, True), strict=False)
    if(not jsn['stderr'] == ''):
        print(jsn['stderr'])
        return JsonResponse({'detail': jsn['stderr']}, status=404)
    
    free_space = float(jsn['stdout'])
    
    return JsonResponse({"size": db_size, "display_size": format_size(db_size), 'free_space': free_space}, status = 200)

@api_view(['POST'])
@permission_classes((IsAuthenticated, ))
def SwitchoverMySQLNode(request):
    """Switching mysql nodes"""
    standby_id = request.data.get("standby_id")
    api_call = request.data.get("rest_api")
    current_user =request.user
    if api_call:
        standby = c123_node.objects.filter(name = standby_id,  type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False).first()
    else:
        standby = c123_node.objects.filter(pk = standby_id,  type = DBType.MYSQL.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False).first()
    if not standby:
        raise APIException410(f"standby id {standby_id} is not available")
    allow_or_403(standby)
    primary = c123_node.objects.filter(cluster = standby.cluster,  type = DBType.MYSQL.value, cluster_role = ClusterRoles.PRIMARY.value, ind_deleted = False).first()
    if not primary:
        raise APIException404(f"primary id {standby_id} is not found")
    in_switchover = Task.objects.filter(task = 'switchover_mysql_node_task', target_id = primary.pk, status__in = [TaskState.PENDING.value, TaskState.RUNNING.value]).first()
    if in_switchover:
        raise APIException406(f"primary {primary.name} are already in switchover (task id {in_switchover.pk})")
    standby.status = Status.SWITCHOVER.value
    primary.status = Status.DEMOTING.value.format(standby.name)
    primary.save()
    standby.save()

    other_standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MYSQL.value, ind_deleted = False, cluster = standby.cluster).exclude(pk = standby.pk)
    standbys = [{"host": object_to_json(i.host), "node": object_to_json(i)}  for i in other_standbys]
    
    tmp_data = json.dumps({'primary_data': object_to_json(primary), 'host_primary' : object_to_json(primary.host), "standbys": standbys, 'standby_data': object_to_json(standby),  'host_standby': object_to_json(standby.host), 'created_by': current_user.id})
    task = Task.objects.create(task='switchover_mysql_node_task',
                                displayment=f'switching {primary.name} to {standby.name}',
                                target_type=DBType.MYSQL.value,
                                target_id=primary.id,
                                org_id = primary.org_id,
                                status=TaskState.PENDING.value,
                                schedule=Schedule.NOW.value,
                                parameters=tmp_data,
                                created_by=current_user)
    return append_status_cmd(request, task)